// ===========================================================================
// Outliner application loader
// ===========================================================================

#include <stdio.h>
#include <dlfcn.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <string.h>
#include <dirent.h>
#include <SDL/SDL.h>
#include "outliner.h"

// ---------------------------------------------------------------------------

#include "cacheflush.c"					// cache flush function for changes made (injected code area)
#include "debug.c"					// debug functions

const char *tagfile = "/mtd_ram/.Outliner";		// tag file for already loaded checking

int delete_debug_log;					// delete debug log after successfull injection ?

const char *IE_BKG_COLOR	= "BKG_COLOR=";		// ini entry for background color
const char *IE_PIXEL_SHIFT	= "PIXEL_SHIFT=";	// ini entry for pixel shift
const char *IE_METHOD		= "METHOD=";		// ini entry for method type
const char *IE_SHADOW_X		= "SHADOW_X=";		// ini entry for x-direction shadow
const char *IE_SHADOW_Y		= "SHADOW_Y=";		// ini entry for y-direction shadow

// ---------------------------------------------------------------------------

// load and process ini file
// returns: 0=error, 1=OK
int load_ini(const char *path, config_struct *ex_config)
{
	FILE *f;
	int i, ln;
	char *p;
	char s[MAX_PATH];

	sprintf(s, "%s%s", path, INI_FILENAME);

	f = fopen(s, "rt");
	if (f)
	{
	   while (!feof(f))
	   {
	      fgets(s, sizeof(s)-1, f);
	      ln = strlen(s);
              if (ln > 0)
              {
		 for (i=0; i<ln; i++)
		     if ((s[i] == 0x0D) || (s[i] == 0x0A)) s[i] = 0;	// cut CR/LF to ASCIIZ
		 ln = strlen(s);
		 if (ln > 0)
		 {
		    if ((s[0]!=';') && (s[0] != '#') && (s[0] != '/'))	// no comment line ?
		    {
			debug("processing ini line:"); debug(s);
			p = strstr(s, IE_BKG_COLOR);	if (p==s) sscanf(p + strlen(IE_BKG_COLOR),	"%lX", &ex_config->bkg_color);
			p = strstr(s, IE_PIXEL_SHIFT);	if (p==s) sscanf(p + strlen(IE_PIXEL_SHIFT),	"%d",  &ex_config->pixel_shift);
			p = strstr(s, IE_METHOD);	if (p==s) sscanf(p + strlen(IE_METHOD),		"%d",  &ex_config->method);
			p = strstr(s, IE_SHADOW_X);	if (p==s) sscanf(p + strlen(IE_SHADOW_X),	"%d",  &ex_config->shadow_x);
			p = strstr(s, IE_SHADOW_Y);	if (p==s) sscanf(p + strlen(IE_SHADOW_Y),	"%d",  &ex_config->shadow_y);
		    }
		 }
	      }
	   }
	   fclose(f);
	}
	return 1; // OK
}

// ---------------------------------------------------------------------------

// injecting procedure
// returns: 1=OK, 0=error
int injector(const char *path)
{
	char s[MAX_PATH];

	debug("injector() started");

	// open self
	unsigned *handle;
	if ( (handle = dlopen(NULL, RTLD_LAZY | RTLD_GLOBAL)) == NULL )
	{
	   debug("cannot open self:");
	   debug(dlerror());
	   return 0;
	}

	debug("self open");

	// getting exeDSP functions...

	// remote control catch functions (press & release)
	unsigned *KeyCommon_SendKeyPressInput = dlsym(handle, "_ZN9KeyCommon17SendKeyPressInputEii");
	if (KeyCommon_SendKeyPressInput == NULL ) return no_symbol(handle, "_ZN9KeyCommon17SendKeyPressInputEii");

	dlclose(handle);


	// Load main library
	sprintf(s, "%s%s", path, MAIN_MODULE);
	debug("loading main module:");
	debug(s);

	if ( (handle = dlopen(s, RTLD_NOW | RTLD_LOCAL | RTLD_NODELETE)) == NULL )
	{
	   debug("cannot open main module:");
	   debug(dlerror());
	   return 0;
	}

	// obtain injection_press procedure address
	unsigned *ex_injection_press = dlsym(handle, "injection_press");
	if (ex_injection_press == NULL) return no_symbol(handle, "injection_press");
	config_struct *ex_config = dlsym(handle, "config");
	if (ex_config == NULL) return no_symbol(handle, "config");

	dlclose(handle);
	debug("main module handle closed");

	// storing app path
	strcpy(ex_config->app_path, path);
	int i = strlen(ex_config->app_path);
	if (i > 0) if (ex_config->app_path[i-1] != '/') strcat(ex_config->app_path, "/");	// add ending slash if not exists
	debug("app_path:"); debug(ex_config->app_path);

	// set other params
	ex_config->bkg_color	= DEF_BKG_COLOR;
	ex_config->pixel_shift	= DEF_PIXEL_SHIFT;
	ex_config->method	= DEF_METHOD;
	ex_config->shadow_x	= DEF_SHADOW_X;
	ex_config->shadow_y	= DEF_SHADOW_Y;

	// load ini config (settings may change here)
	if (!load_ini(path, ex_config)) return 0;

	// Injecting!
 	int page = ((unsigned) KeyCommon_SendKeyPressInput) & ~0x0FFF;		// it seems that it is enough to unlock this page only
	mprotect((void *) page, 8192, PROT_READ | PROT_WRITE | PROT_EXEC);
	if (KeyCommon_SendKeyPressInput[4] == 0xE51FF004)	// remapped already (by other application) ?
	{
	   debug("old injection_press detected");

	   // modify ex_injection_press returning command (self-injection)
	   int page2 = ((unsigned) ex_injection_press) & ~0x0FFF;
	   mprotect((void *) page2, 8192, PROT_READ | PROT_WRITE | PROT_EXEC);
	   ex_injection_press[9]  = KeyCommon_SendKeyPressInput[4];
	   ex_injection_press[10] = KeyCommon_SendKeyPressInput[5];
	   mprotect((void *) page2, 8192, PROT_READ | PROT_EXEC);
	   cacheflush(ex_injection_press+9, 8);
	}
	else debug ("first injection_press");

	// set "classic" injection_press function
	KeyCommon_SendKeyPressInput[4] = 0xE51FF004;
	KeyCommon_SendKeyPressInput[5] = (unsigned)ex_injection_press;

	mprotect((void *) page, 8192, PROT_READ | PROT_EXEC);

	cacheflush(KeyCommon_SendKeyPressInput+4, 8);

	debug("Injection complete");
	return 1;
}

// ---------------------------------------------------------------------------

// check if our module has been already loaded and attached to exeDSP process
int already_loaded()
{
   // simple check by tag file
   if (access(tagfile, F_OK) == 0) return 1;	// OK: tag file present => application already loaded
   return 0;					// OK: module not loaded
}

// ---------------------------------------------------------------------------

// main plugin ("Game") function
int Game_Main(const char *path, const char *udn __attribute__ ((unused)))
{
	FILE *f;

	delete_debug_log = 1;

	if (!already_loaded())
	{
	   if (injector(path))		// load, configure and inject our code leaving it in memory
	   {
	      if (delete_debug_log) remove(DEBUG_LOG);	// all went OK, so we can delete log file

	      // application has been loaded so we must set tagfile to prevent against multiloading
	      f = fopen(tagfile, "wt"); if (f) fclose(f);
	   }

	   return 1;	// exit loader while injected module stays in memory
	}
	else
	{
	   debug("Outliner already installed");
	}

	return 1;
}

// ---------------------------------------------------------------------------

