// ===========================================================================
// Outliner 0.2 application
// (c)2011-05-21...24 by geo650 / SamyGO
// ===========================================================================

#include <stdio.h>
#include <pthread.h>
#include <SDL/SDL.h>

#include "outliner.h"			// app definitions
#include "buttons.h"			// button codes

// ---------------------------------------------------------------------------

int outliner_enabled = 0;		// is outliner engine enabled?
int allow_thread = 0;			// allow thread execution?
pthread_t pthr;				// thread handle
int thread_ID;				// thread ID
int threading;				// threading flag
int checksum, last_checksum;		// last subtitle checksum to test if changed
SDL_Surface *screen = NULL;		// screen surface handle
config_struct config;			// configuration structure

// ---------------------------------------------------------------------------

// clear OSD screen (set all RGBA pixels to 0)
void OSD_cls(int flip)
{
   unsigned char *scr;
   int y;

   if (screen)
   {
      SDL_LockSurface(screen);
      scr = screen->pixels;
      for (y = 0; y < SCREEN_H; y++)
      {
	  memset(scr, 0, SCREEN_W * SCREEN_BPP/8);
	  scr += screen->pitch;	// go to the next line
      }
      SDL_UnlockSurface(screen);
      if (flip) SDL_Flip(screen);
   }
}

// ---------------------------------------------------------------------------

#define SNAPSHOT_SIZE	(SCREEN_W * (SCREEN_H - MARG_UP - MARG_DOWN))		// snapshot size [pixels]

// shadow method - main function
void shadow()
{
	int i;
	Uint32 *src, *dst;							// source and destination surface pixels' addresses
	static Uint32 buffer[SNAPSHOT_SIZE * 4];				// surface buffer

	if (!outliner_enabled) return;
	if (threading) return;
	threading = 1;								// thread's step is starting

	// make snapshot
	src = (Uint32 *)(SRC_ADDR + config.pixel_shift + (MARG_UP * SCREEN_W * 4));	// source address
	memcpy(&buffer, src, SNAPSHOT_SIZE * 4);

	// edit snapshot
	for (i=0; i < SNAPSHOT_SIZE; i++) buffer[i] &= (Uint32)0xFF000000;	// make text black (leave alpha-channel only)

	// place edited snapshot on the screen surface (under subtitles)
	SDL_LockSurface(screen);						// lock surface

	// address will be shifted to make a shadow
	dst = (Uint32 *)(screen->pixels + screen->pitch * (MARG_UP + config.shadow_y) + 4 * config.shadow_x);

	memcpy(dst, &buffer, SNAPSHOT_SIZE * 4);

	SDL_UnlockSurface(screen);						// unlock surface
	SDL_Flip(screen);							// flip (display) surface

	threading = 0;								// end of thread's step
}

// ---------------------------------------------------------------------------

// bars method - main function
void bars()
{
	SDL_Rect row[MAX_BARS];
	int bars, is_empty;
	int x, y, w, px;
	Uint32 *src, *dst;

	if (!outliner_enabled) return;
	if (threading) return;
	threading = 1;								// thread's step is starting

	// find subtitle rows (i.e. non-transparent pixel areas of SDL subtitle surface)
	bars = 0;
	checksum = 0;
	is_empty = 1;
	for (y = MARG_UP; y < (SCREEN_H - MARG_DOWN); y++)
	{
	    src = (Uint32 *)(SRC_ADDR + config.pixel_shift + (y * SCREEN_W * 4) + (MARG_LEFT * 4));	// source address of 'y' screen line shifted by MARG_LEFT pixels right
	    w = SCREEN_W - MARG_LEFT - MARG_RIGHT;				// number of pixels to examine in this line

	    px = MARG_LEFT;
	    while ((*src == BLACK_COLOR) && (w--)) { src++; px++; }		// scan for first non-zero pixel

	    if (w > 0)								// found ?
	    {
	       if (is_empty)							// first found line after "emptiness" ?
	       {
		  is_empty = 0;							// this line is not empty
		  row[bars].y = y - MARG_HBAR;					// add this line to the bar
		  if (row[bars].y < 0) row[bars].y = 0;				// and add its upper margin as well
		  row[bars].h = y - row[bars].y + 1;				// and then set initial height (==1+MARG_HBAR, limited at top)
		  row[bars].x = px;						// init bar's x position

		  // find end of non-empty scanline (line width)
		  src = (Uint32 *)(SRC_ADDR + config.pixel_shift + ((y+1) * SCREEN_W * 4) - ((MARG_RIGHT+1) * 4));	// source address of 'y' screen line's end with right margin
		  w = SCREEN_W - MARG_LEFT - MARG_RIGHT;					// number of pixels to examine in this line

		  px = 0;
		  while ((*src == BLACK_COLOR) && (w--)) { src--; px++;	}	// count zero-pixels

		  row[bars].w = SCREEN_W - MARG_RIGHT - px - row[bars].x;	// initial non-empty scanline width

		  bars++;							// and increase bars number
		  checksum += px + SCREEN_H * y;				// simple checksum (TODO: maybe it would be nice to improve this "algorithm")
	       }
	       else
	       {
		  row[bars-1].h++;						// increase bar height
		  if (row[bars-1].x > px) row[bars-1].x = px;			// set minimum x position of bar

		  // find end of non-empty scanline (line width) again
		  src = (Uint32 *)(SRC_ADDR + config.pixel_shift + ((y+1) * SCREEN_W * 4) - ((MARG_RIGHT+1) * 4));	// source address of 'y' screen line's end with right margin
		  w = SCREEN_W - MARG_LEFT - MARG_RIGHT;					// number of pixels to examine in this line

		  px = 0;
		  while ((*src == BLACK_COLOR) && (w--)) { src--; px++;	}			// count zero-pixels

		  if (row[bars-1].w < (SCREEN_W - MARG_RIGHT - px - row[bars-1].x))		// fix bar width
		      row[bars-1].w =  SCREEN_W - MARG_RIGHT - px - row[bars-1].x;		//
	       }
	    }
	    else								// empty line ?
	    {
	       if (is_empty == 0)						// and we have at least one row (line was not empty before) ?
	       {
		  row[bars-1].h += MARG_HBAR;					// add lower bar margin
		  if (row[bars-1].y + row[bars-1].h >= SCREEN_H)		// and limit it
		     row[bars-1].h = SCREEN_H - row[bars-1].y - 1; 		// to the screen height
		  is_empty = 1;							// this scanline was empty

		  checksum += row[bars-1].w;					// additional step for simple checksum (TODO: see above)

		  // check for bars limit number
		  if (bars+1 >= MAX_BARS) y = SCREEN_H - MARG_DOWN; 		// limit reached! so please exit y-loop
	       }
	    }
	}

	// do we need to draw something?
	if (checksum == last_checksum) { threading = 0; return;	}		// no, we don't
	last_checksum = checksum;						// we do, so remember this checksum now

	// draw bars
	OSD_cls(0);								// first, clear screen
	SDL_LockSurface(screen);						// then lock surface

	while (--bars >= 0)
	{
	   // add extra columns
	   row[bars].x -= MARG_WBAR; if (row[bars].x < 0) row[bars].x = 0;
	   row[bars].w += 2*MARG_WBAR; if (row[bars].x + row[bars].w >= SCREEN_W) row[bars].w = SCREEN_H - row[bars].x - 1;

	   dst = (Uint32 *)(screen->pixels + (screen->pitch * row[bars].y));	// destination address of upper bar's scanline
	   for (y = 0; y < row[bars].h; y++)
	   {
	       dst += row[bars].x;						// shift x pixels (skip left pixels)
	       x = row[bars].w;
	       while (x--) *(dst++) = config.bkg_color;				// draw background pixels
	       dst += (screen->pitch >> 2) - row[bars].w - row[bars].x;		// shift to the end of scanline (skip right pixels)
	   }
	}

	SDL_UnlockSurface(screen);						// unlock surface
	SDL_Flip(screen);							// flip (display) surface

	threading = 0;								// end of thread's step
}

// ---------------------------------------------------------------------------

// outliner thread (working in endless loop until stopped)
void *outliner_thread(void *ptr)
{
	while (allow_thread)
	{
	   if (outliner_enabled)
	   {
	      if (config.method == METHOD_BARS) bars();		// call main bars function
	      else
	      if (config.method == METHOD_SHADOW) shadow();	// call main shadow function
	   }
	   else { allow_thread=0; OSD_cls(1); return ptr; }	// not enabled!
	   SDL_Delay(CPU_WAIT);					// drop CPU usage
	}
	OSD_cls(1);						// end of thread? clear existing bars
	return ptr;						// end of thread
}

// ---------------------------------------------------------------------------

// enable outliner engine
void enable_outliner()
{
	if (outliner_enabled) return;				// enabled already

	last_checksum = -1;

	if (!screen) SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER);

	// set video mode for bars
	if (!screen) screen = SDL_SetVideoMode(SCREEN_W, SCREEN_H, SCREEN_BPP, SCREEN_FLAGS);
	if (!screen)
	{
	   //SDL_Quit();
	   outliner_enabled = 0;				// cannot set video mode!
	}
	else
	{
	   outliner_enabled = 1;				// OK

	   // start thread
	   if (allow_thread) return;				// thread already created and working
	   // thread not created or exited...
	   allow_thread = 1;
	   threading = 0;
	   thread_ID = pthread_create(&pthr, NULL, &outliner_thread, NULL);
	}
}

// ---------------------------------------------------------------------------

// disable outliner engine
void disable_outliner()
{
	if (outliner_enabled)
	{
	   allow_thread = 0;
	   SDL_Delay(500);					// give a chance to end our thread
	   //SDL_Quit();
	   //screen = NULL;
	   outliner_enabled = 0;
	}
}

// ---------------------------------------------------------------------------

// called at keypress
static int button_pressed(int key) __attribute__ ((noinline));
static int button_pressed(int key)
{
	switch(key)
	{
	   case KEY_PLAY:
		enable_outliner();
		//key = KEY_NOTHING;
		break;
	   case KEY_STOP:
	   case KEY_RETURN:
	   case KEY_EXIT:
	   case KEY_TV:
		disable_outliner();
		break;

	   default:
		;
	}

	return(key);				// return destination button code to be processed by exeDSP
}


// ---------------------------------------------------------------------------

// injection procedure for press
void injection_press(int a, int key, int arg_r2) __attribute__ ((noinline));
void injection_press(int a, int key, int arg_r2)
{
	// CAUTION! Do not add/modify code here... or modify injection indexes [9][10] in loader

	// [0][1] gcc should put "STMFD SP!, {R4,LR}" and "MOV R4,[R1]" here
	asm volatile(
		"STMFD   SP!, {R0,R2,R3,R12}\n"		// [2] store some registers
	);

	// first, we have to execute our function
	button_pressed(key);				// [3][4] calling local function with key variable in R0, returning R0

	// then, prepare for calling and call previous injection function
	asm volatile(
		"MOV     R1, R0\n"			// [5] remapped keycode
		"MOV     R6, R0\n"			// [6] remapped keycode
		"LDMFD   SP!, {R0,R2,R3,R12}\n"		// [7] restore registers stored before
		"LDMFD   SP!, {R4,LR}\n"		// [8] restore registers...
		"MOV     R7, R2\n"			// [9] This is original command that we overwrite (and replace at next injection)
		"LDR     R3, [R8]\n"			// [10] This is original command that we overwrite (and replace at next injection)
		"LDR     PC, =(_ZN9KeyCommon17SendKeyPressInputEii + 0x18 )\n"	// [11][12] return after injection (not reached at next injection)
	);
	// and that's all; the rest (returning to exeDSP) is already programmed in the first injected function
}

// ---------------------------------------------------------------------------
// end.

